"""Template langage for Rougail to create file and systemd file

Cadoles (http://www.cadoles.com)
Copyright (C) 2021

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from typing import Dict, Any
from os import makedirs, symlink
from os.path import dirname, isfile, join
from ipaddress import ip_network

from .base import RougailBaseTemplate
from ..i18n import _
from ..error import FileNotFound


ROUGAIL_IP_TEMPLATE = """[Service]
%for %%ip in %%rougail_variable
IPAddressAllow=%%ip
%end for
IPAddressDeny=any
"""


ROUGAIL_TMPL_TEMPLATE = """%def display(%%file, %%filename)
%if %%filename.startswith('/etc/') or %%filename.startswith('/var/') or %%filename.startswith('/srv/')
C %%filename %%file.mode %%file.owner %%file.group - /usr/local/lib%%filename
z %%filename - - - - -
%end if
%end def
%for %%service in %%services
  %if %%service.activate is True and %%hasattr(%%service, 'files')
    %for %%file in %%service.files
      %if %%file.activate is True and %%file.included != 'content'
        %if %%isinstance(%%file.name, list)
          %for %%filename in %%file.name
%%display(%%file, %%filename)%slurp
          %end for
        %else
%%display(%%file, %%file.name)%slurp
        %end if
      %end if
    %end for
  %end if
%end for"""


class RougailSystemdTemplate(RougailBaseTemplate):
    def __init__(self,  # pylint: disable=R0913
                 config: 'Config',
                 rougailconfig: 'RougailConfig'=None,
                 ) -> None:
        self.ip_per_service = None
        super().__init__(config, rougailconfig)

    def _instance_files(self,
                        filevar: Dict,
                        destfile: str,
                        service_name: str,
                        variable,
                        idx: int,
                        ) -> tuple:
        source = filevar['source']
        if not isfile(source):  # pragma: no cover
            raise FileNotFound(_(f"File {source} does not exist."))
        tmp_file = join(self.tmp_dir, source)
        #self.instance_file(fill, 'files')
        if variable:
            var = variable[idx]
        else:
            var = None
        return tmp_file, None, destfile, var

    def _instance_overrides(self,
                            filevar: Dict,
                            destfile,
                            service_name: str,
                            *args,
                            ) -> tuple:
        source = filevar['source']
        if not isfile(source):  # pragma: no cover
            raise FileNotFound(_(f"File {source} does not exist."))
        tmp_file = join(self.tmp_dir, source)
        service_name = filevar['name']
        return tmp_file, None, f'/systemd/system/{service_name}.service.d/rougail.conf', None

    def _instance_ip(self,
                     filevar: Dict,
                     ip,
                     service_name: str,
                     var: Any,
                     idx: int,
                     *args,
                     ) -> tuple:
        if self.ip_per_service is None:
            self.ip_per_service = []
        if 'netmask' in filevar:
            if isinstance(filevar["netmask"], list):
                netmask = filevar['netmask'][idx]
            else:
                netmask = filevar['netmask']
            self.ip_per_service.append(str(ip_network(f'{ip}/{netmask}')))
        elif ip:
            self.ip_per_service.append(ip)

    def desactive_service(self,
                          service_name: str,
                          ):
        filename = f'{self.destinations_dir}/systemd/system/{service_name}.service'
        makedirs(dirname(filename), exist_ok=True)
        symlink('/dev/null', filename)

    def post_instance_service(self,
                              service_name: str,
                              ) -> None:  # pragma: no cover
        if self.ip_per_service is None:
            return
        destfile = f'/systemd/system/{service_name}.service.d/rougail_ip.conf'
        destfilename = join(self.destinations_dir, destfile[1:])
        makedirs(dirname(destfilename), exist_ok=True)
        self.log.info(_(f"creole processing: '{destfilename}'"))
        self.engines['creole'].process(filename=None,
                                       source=ROUGAIL_IP_TEMPLATE,
                                       true_destfilename=destfile,
                                       destfilename=destfilename,
                                       destdir=self.destinations_dir,
                                       variable=self.ip_per_service,
                                       rougail_variables_dict=self.rougail_variables_dict,
                                       eosfunc=self.eosfunc,
                                       )
        self.ip_per_service = None

    def post_instance(self):
        destfile = '/tmpfiles.d/rougail.conf'
        destfilename = join(self.destinations_dir, destfile[1:])
        makedirs(dirname(destfilename), exist_ok=True)
        self.log.info(_(f"creole processing: '{destfilename}'"))
        self.engines['creole'].process(filename=None,
                                       source=ROUGAIL_TMPL_TEMPLATE,
                                       true_destfilename=destfile,
                                       destfilename=destfilename,
                                       destdir=self.destinations_dir,
                                       variable=None,
                                       rougail_variables_dict=self.rougail_variables_dict,
                                       eosfunc=self.eosfunc,
                                       )
